from django.conf import settings
from django.db import models, transaction

from django_pgschemas.models import TenantModel
from django_pgschemas.utils import get_domain_model


class DomainModel(models.Model):
    """
    All models that store the domains must inherit this class.
    """

    tenant = (
        models.ForeignKey(
            settings.TENANTS["default"]["TENANT_MODEL"],
            db_index=True,
            related_name="domains",
            on_delete=models.CASCADE,
        )
        if getattr(settings, "TENANTS", {}).get("default")
        else None
    )

    domain = models.CharField(max_length=253, db_index=True)
    folder = models.SlugField(max_length=253, blank=True, db_index=True)

    is_primary = models.BooleanField(default=True)
    redirect_to_primary = models.BooleanField(default=False)

    class Meta:
        abstract = True
        unique_together = (("domain", "folder"),)

    def __str__(self) -> str:
        return "/".join([self.domain, self.folder]) if self.folder else self.domain

    @transaction.atomic
    def save(self, *args: object, **kwargs: object) -> None:
        using = kwargs.get("using")
        domain_list = self.__class__.objects
        if using:
            domain_list = domain_list.using(using)
        domain_list = domain_list.filter(tenant=self.tenant, is_primary=True).exclude(pk=self.pk)
        self.is_primary = self.is_primary or (not domain_list.exists())
        if self.is_primary:
            domain_list.update(is_primary=False)
            if self.redirect_to_primary:
                self.redirect_to_primary = False
        super().save(*args, **kwargs)

    def absolute_url(self, path: str) -> str:
        """
        Constructs an absolute url for this domain / folder and a given path
        """
        parts = [self.domain]

        if self.folder:
            parts.append(self.folder)

        parts.append(path)

        final_path = "/".join(parts).replace("//", "/")

        return f"//{final_path}"


def get_primary_domain_for_tenant(tenant: TenantModel) -> DomainModel | None:
    DomainModel = get_domain_model()

    if DomainModel is None:
        return None

    try:
        return tenant.domains.get(is_primary=True)
    except DomainModel.DoesNotExist:
        return None



from django_pgschemas.schema import Schema

from . import SchemaCommand


class Command(SchemaCommand):
    help = "Displays which schemas would be used based on the passed schema selectors"

    def handle_schema(self, schema: Schema, *args, **options):
        if options["verbosity"] >= 1:
            self.stdout.write(str(schema.routing) if schema.routing else schema.schema_name)



from importlib import import_module
from types import ModuleType
from typing import Callable

from django.conf import settings
from django.db import DEFAULT_DB_ALIAS

DEFAULT_BACKEND = "django.db.backends.postgresql"


def get_tenant_db_alias() -> str:
    return getattr(settings, "PGSCHEMAS_TENANT_DB_ALIAS", DEFAULT_DB_ALIAS)


def get_limit_set_calls() -> bool:
    return getattr(settings, "PGSCHEMAS_LIMIT_SET_CALLS", False)


def get_original_backend() -> str:
    return getattr(settings, "PGSCHEMAS_ORIGINAL_BACKEND", DEFAULT_BACKEND)


def get_extra_search_paths() -> list[str]:
    return getattr(settings, "PGSCHEMAS_EXTRA_SEARCH_PATHS", [])


def get_tenant_session_key() -> str:
    return getattr(settings, "PGSCHEMAS_TENANT_SESSION_KEY", "tenant")


def get_tenant_header() -> str:
    return getattr(settings, "PGSCHEMAS_TENANT_HEADER", "tenant")


def get_pathname_function() -> Callable | None:
    return getattr(settings, "PGSCHEMAS_PATHNAME_FUNCTION", None)


def get_base_backend_module(submodule: str | None = None) -> ModuleType:
    module = DEFAULT_BACKEND
    if submodule:
        module += f".{submodule}"
    return import_module(module)


def get_original_backend_module(submodule: str | None = None) -> ModuleType:
    module = get_original_backend()
    if submodule:
        module += f".{submodule}"
    return import_module(module)



import re
import sys
from types import ModuleType
from typing import Any, Literal

from django.conf import settings
from django.urls import URLResolver

from django_pgschemas.routing.info import DomainInfo, HeadersInfo, SessionInfo
from django_pgschemas.schema import Schema, get_current_schema

DYNAMIC_URLCONF_SUFFIX = "_dynamically_tenant_prefixed"


class TenantPrefixPattern:
    converters: dict = {}

    @property
    def tenant_prefix(self) -> str:
        current_schema = get_current_schema()
        return (
            f"{current_schema.routing.folder}/"
            if isinstance(current_schema.routing, DomainInfo) and current_schema.routing.folder
            else ""
        )

    @property
    def regex(self) -> re.Pattern:
        # This is only used by reverse() and cached in _reverse_dict.
        return re.compile(self.tenant_prefix)

    def match(self, path: str) -> tuple | None:
        tenant_prefix = self.tenant_prefix
        if path.startswith(tenant_prefix):
            return path[len(tenant_prefix) :], (), {}
        return None

    def check(self) -> list:
        return []

    def describe(self) -> str:
        return f"'{self}'"

    def __str__(self) -> str:
        return self.tenant_prefix


def get_dynamic_tenant_prefixed_urlconf(urlconf: str, dynamic_path: str) -> ModuleType:
    """
    Generates a new urlconf module with all patterns prefixed with tenant.
    """

    class LazyURLConfModule(ModuleType):
        def __getattr__(self, attr: str) -> Any:
            if attr == "urlpatterns":
                return [URLResolver(TenantPrefixPattern(), urlconf)]
            return self.__getattribute__(attr)

    return LazyURLConfModule(dynamic_path)


def _get_urlconf_from_schema(
    schema: Schema, config_key: Literal["URLCONF", "WS_URLCONF"]
) -> str | None:
    match schema.routing:
        case DomainInfo(domain, _):
            # Checking for static tenants
            if not schema.is_dynamic:
                for schema_name, data in settings.TENANTS.items():
                    if schema_name in ["public", "default"]:
                        continue
                    if domain in data.get("DOMAINS", []):
                        return data.get(config_key)
                    if domain in data.get("FALLBACK_DOMAINS", []):
                        return data.get(config_key)
                return None

            # Checking for dynamic tenants
            urlconf = settings.TENANTS.get("default", {}).get(config_key)
            if urlconf is not None and schema.routing.folder:
                dynamic_path = urlconf + DYNAMIC_URLCONF_SUFFIX
                if not sys.modules.get(dynamic_path):
                    sys.modules[dynamic_path] = get_dynamic_tenant_prefixed_urlconf(
                        urlconf, dynamic_path
                    )
                urlconf = dynamic_path

            return urlconf

        case SessionInfo(reference):
            # Checking for static tenants
            if not schema.is_dynamic:
                for schema_name, data in settings.TENANTS.items():
                    if schema_name in ["public", "default"]:
                        continue
                    if reference == data.get("SESSION_KEY"):
                        return data.get(config_key)
                return None

            # Checking for dynamic tenants
            return settings.TENANTS.get("default", {}).get(config_key)

        case HeadersInfo(reference):
            # Checking for static tenants
            if not schema.is_dynamic:
                for schema_name, data in settings.TENANTS.items():
                    if schema_name in ["public", "default"]:
                        continue
                    if reference == data.get("HEADER"):
                        return data.get(config_key)
                return None

            # Checking for dynamic tenants
            return settings.TENANTS.get("default", {}).get(config_key)

        case _:
            return None


def get_urlconf_from_schema(schema: Schema) -> str | None:
    return _get_urlconf_from_schema(schema, "URLCONF")


def get_ws_urlconf_from_schema(schema: Schema) -> str | None:
    return _get_urlconf_from_schema(schema, "WS_URLCONF")



from dataclasses import dataclass
from typing import TypeAlias


@dataclass(frozen=True)
class DomainInfo:
    domain: str
    folder: str | None = None

    def __str__(self) -> str:
        return f"{self.domain}/{self.folder}" if self.folder else self.domain


@dataclass(frozen=True)
class SessionInfo:
    reference: str

    def __str__(self) -> str:
        return f"Session: {self.reference}"


@dataclass(frozen=True)
class HeadersInfo:
    reference: str

    def __str__(self) -> str:
        return f"Header: {self.reference}"


RoutingInfo: TypeAlias = DomainInfo | SessionInfo | HeadersInfo | None



import os

from django.core.files.storage import FileSystemStorage

from django_pgschemas.routing.info import DomainInfo
from django_pgschemas.schema import get_current_schema
from django_pgschemas.settings import get_pathname_function


class TenantFileSystemStorage(FileSystemStorage):
    """
    Tenant aware file system storage. Appends the tenant identifier to the base
    location and base URL.
    """

    def get_schema_path_identifier(self):
        schema = get_current_schema()

        if schema is None:
            return ""

        path_identifier = schema.schema_name

        if hasattr(schema, "schema_pathname"):
            path_identifier = schema.schema_pathname()

        elif pathname_function := get_pathname_function():
            path_identifier = pathname_function(schema)

        return path_identifier

    @property  # To avoid caching of tenant
    def base_location(self):
        """
        Appends base location with the schema path identifier.
        """
        file_folder = self.get_schema_path_identifier()
        location = os.path.join(super().base_location, file_folder)

        if not location.endswith("/"):
            location += "/"

        return location

    @property  # To avoid caching of tenant
    def location(self):
        return super().location

    @property  # To avoid caching of tenant
    def base_url(self):
        """
        Optionally appends base URL with the schema path identifier.
        If the current schema is already using a folder, no path identifier is
        appended.
        """
        schema = get_current_schema()
        url_folder = self.get_schema_path_identifier()

        # Specific case of domain+folder routing
        if (
            url_folder
            and schema
            and isinstance(schema.routing, DomainInfo)
            and schema.routing.folder
        ):
            # Since we're already prepending all URLs with schema, there is no
            # need to make the differentiation here
            url_folder = ""

        parent_base_url = super().base_url.strip("/")
        url = "/".join(["", parent_base_url, url_folder])

        if not url.endswith("/"):
            url += "/"

        return url



from django.db import models

from django_pgschemas.postgresql.base import check_schema_name
from django_pgschemas.schema import Schema
from django_pgschemas.signals import (
    dynamic_tenant_needs_sync,
    dynamic_tenant_post_sync,
    dynamic_tenant_pre_drop,
)
from django_pgschemas.utils import (
    create_or_clone_schema,
    drop_schema,
    schema_exists,
)


class TenantModel(Schema, models.Model):
    """
    All tenant models must inherit this class.
    """

    auto_create_schema = True
    """
    Set this flag to `False` on a parent class if you don't want the schema
    to be automatically created upon save.
    """

    auto_drop_schema = False
    """
    *USE THIS WITH CAUTION!*
    Set this flag to `True` on a parent class if you want the schema to be
    automatically deleted if the tenant row gets deleted.
    """

    is_dynamic = True
    """
    Leave this as `True`. Denotes it's a database controlled tenant.
    """

    schema_name = models.CharField(max_length=63, unique=True, validators=[check_schema_name])

    class Meta:
        abstract = True

    def save(
        self,
        force_insert: bool = False,
        force_update: bool = False,
        using: str | None = None,
        update_fields: list[str] | None = None,
        verbosity: int = 1,
    ) -> None:
        is_new = self.pk is None

        super().save(force_insert, force_update, using, update_fields)

        if is_new and self.auto_create_schema:
            try:
                self.create_schema(verbosity=verbosity)
                dynamic_tenant_post_sync.send(sender=TenantModel, tenant=self.serializable_fields())
            except Exception:
                # We failed creating the tenant, delete what we created and re-raise the exception
                self.delete(force_drop=True)
                raise
        elif is_new:
            # Although we are not using the schema functions directly, the signal might be registered by a listener
            dynamic_tenant_needs_sync.send(sender=TenantModel, tenant=self.serializable_fields())
        elif not is_new and self.auto_create_schema and not schema_exists(self.schema_name):
            # Create schemas for existing models, deleting only the schema on failure
            try:
                self.create_schema(verbosity=verbosity)
                dynamic_tenant_post_sync.send(sender=TenantModel, tenant=self.serializable_fields())
            except Exception:
                # We failed creating the schema, delete what we created and re-raise the exception
                self.drop_schema()
                raise

    def delete(
        self, using: str | None = None, keep_parents: bool = False, force_drop: bool = False
    ) -> None:
        """
        Deletes this row. Drops the tenant's schema if the attribute
        `auto_drop_schema` is `True`.
        """
        if force_drop or self.auto_drop_schema:
            dynamic_tenant_pre_drop.send(sender=TenantModel, tenant=self.serializable_fields())
            self.drop_schema()

        super().delete(using, keep_parents)

    def serializable_fields(self) -> "TenantModel":
        """
        In certain cases the model isn't serializable so you may want to only
        send the id.
        """
        return self

    def create_schema(self, sync_schema: bool = True, verbosity: int = 1) -> bool:
        """
        Creates or clones the schema `schema_name` for this tenant.
        """
        return create_or_clone_schema(self.schema_name, sync_schema, verbosity)

    def drop_schema(self) -> bool:
        """
        Drops the schema.
        """
        return drop_schema(self.schema_name)



from contextlib import contextmanager
from contextvars import ContextVar, Token
from functools import lru_cache
from typing import Iterator

from django_pgschemas.routing.info import RoutingInfo
from django_pgschemas.signals import schema_activate


class Schema:
    schema_name: str
    routing: RoutingInfo = None

    is_dynamic = False

    _context_tokens: list[Token["Schema"] | None] = []

    @staticmethod
    def create(schema_name: str, routing: RoutingInfo | None = None) -> "Schema":
        schema = Schema()
        schema.schema_name = schema_name
        schema.routing = routing
        return schema

    def __enter__(self) -> None:
        self._context_tokens.append(push(self))

    def __exit__(self, *args: object) -> None:
        if self._context_tokens:
            token = self._context_tokens.pop()
            if token is not None:
                active.reset(token)


def shallow_equal(schema1: Schema, schema2: Schema) -> bool:
    return schema1.schema_name == schema2.schema_name and schema1.routing == schema2.routing


@lru_cache
def get_default_schema() -> Schema:
    return Schema.create("public")


active: ContextVar["Schema"] = ContextVar("active", default=get_default_schema())


def get_current_schema() -> Schema:
    return active.get()


def push(schema: Schema) -> Token[Schema] | None:
    if shallow_equal(get_current_schema(), schema):
        return None

    token = active.set(schema)

    schema_activate.send(sender=Schema, schema=schema)

    return token


def activate(schema: Schema) -> None:
    push(schema)


def deactivate() -> None:
    push(get_default_schema())


activate_public = deactivate


@contextmanager
def override(schema: Schema) -> Iterator[None]:
    token = push(schema)

    yield

    if token is not None:
        active.reset(token)



from typing import cast

from channels.db import database_sync_to_async
from channels.middleware import BaseMiddleware
from channels.routing import URLRouter
from django.conf import settings
from django.db.models import Q
from django.urls import URLResolver, path
from django.utils.encoding import force_bytes, force_str
from django.utils.module_loading import import_string

from django_pgschemas.models import TenantModel as TenantModelBase
from django_pgschemas.routing.info import DomainInfo, HeadersInfo
from django_pgschemas.routing.urlresolvers import get_ws_urlconf_from_schema
from django_pgschemas.schema import Schema
from django_pgschemas.settings import get_tenant_header
from django_pgschemas.utils import get_domain_model, get_tenant_model, remove_www


def TenantURLRouter():
    async def router(scope, receive, send):
        schema: Schema | None = scope.get("tenant")
        routes = []

        if schema is not None:
            ws_urlconf = get_ws_urlconf_from_schema(schema) if schema else None

            if ws_urlconf:
                routes = import_string(ws_urlconf + ".urlpatterns")
                match (routes, schema.routing):
                    case ([URLResolver()], DomainInfo(_, folder)) if folder:
                        routes = [path(f"{folder}/", URLRouter(routes[0].url_patterns))]
                    case _:
                        pass

        _router = URLRouter(routes)
        return await _router(scope, receive, send)

    return router


class BaseRoutingMiddleware(BaseMiddleware):
    @database_sync_to_async
    def get_scope_tenant(self, scope):
        raise NotImplementedError

    async def __call__(self, scope, receive, send):
        scope = dict(scope)
        scope["tenant"] = await self.get_scope_tenant(scope)

        return await super().__call__(scope, receive, send)


class DomainRoutingMiddleware(BaseRoutingMiddleware):
    @database_sync_to_async
    def get_scope_tenant(self, scope) -> Schema | None:
        hostname = force_str(dict(scope["headers"]).get(b"host", b""))
        hostname = remove_www(hostname.split(":")[0])

        tenant: Schema | None = None

        # Checking for static tenants
        for schema, data in settings.TENANTS.items():
            if schema in ["public", "default"]:
                continue
            if hostname in data.get("DOMAINS", []):
                tenant = Schema.create(
                    schema_name=schema,
                    routing=DomainInfo(domain=hostname),
                )
                break

        # Checking for dynamic tenants
        else:
            DomainModel = get_domain_model()

            prefix = scope["path"].split("/")[1]
            domain = None

            if DomainModel is not None:
                try:
                    domain = DomainModel.objects.select_related("tenant").get(
                        domain=hostname, folder=prefix
                    )
                except DomainModel.DoesNotExist:
                    try:
                        domain = DomainModel.objects.select_related("tenant").get(
                            domain=hostname, folder=""
                        )
                    except DomainModel.DoesNotExist:
                        pass

            if domain is not None:
                tenant = cast(TenantModelBase, domain.tenant)
                tenant.routing = DomainInfo(domain=hostname)

                if prefix and domain.folder == prefix:
                    tenant.routing = DomainInfo(domain=hostname, folder=prefix)

        # Checking fallback domains
        if not tenant:
            for schema, data in settings.TENANTS.items():
                if schema in ["public", "default"]:
                    continue
                if hostname in data.get("FALLBACK_DOMAINS", []):
                    tenant = Schema.create(
                        schema_name=schema,
                        routing=DomainInfo(domain=hostname),
                    )
                    break

        return tenant


class HeadersRoutingMiddleware(BaseRoutingMiddleware):
    @database_sync_to_async
    def get_scope_tenant(self, scope) -> Schema | None:
        tenant_header = get_tenant_header()
        tenant_ref = force_str(dict(scope["headers"]).get(force_bytes(tenant_header), b""))

        if not tenant_ref:
            return None

        tenant: Schema | None = None

        # Checking for static tenants
        for schema, data in settings.TENANTS.items():
            if schema in ["public", "default"]:
                continue
            if tenant_ref == schema or tenant_ref == data.get("HEADER"):
                tenant = Schema.create(
                    schema_name=schema, routing=HeadersInfo(reference=tenant_ref)
                )
                break

        # Checking for dynamic tenants
        else:
            if (TenantModel := get_tenant_model()) is not None:
                tenant = TenantModel._default_manager.filter(
                    Q(pk__iexact=tenant_ref) | Q(schema_name=tenant_ref)
                ).first()
                if tenant is not None:
                    tenant.routing = HeadersInfo(reference=tenant_ref)

        return tenant



from typing import Any

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.sessions.base_session import AbstractBaseSession
from django.core import checks
from django.core.exceptions import ImproperlyConfigured
from django.db import connection
from django.db.utils import ProgrammingError
from django.utils.module_loading import import_module

from django_pgschemas.settings import get_extra_search_paths
from django_pgschemas.utils import (
    get_clone_reference,
    get_domain_model,
    get_tenant_model,
    is_valid_schema_name,
)


def get_tenant_app() -> str | None:
    model = get_tenant_model()

    if model is None:
        return None

    return model._meta.app_config.name


def get_domain_app() -> str | None:
    model = get_domain_model()

    if model is None:
        return None

    return model._meta.app_config.name


def get_user_app() -> str | None:
    try:
        return get_user_model()._meta.app_config.name
    except (AttributeError, ImproperlyConfigured):
        return None


def get_session_app() -> str | None:
    engine = import_module(settings.SESSION_ENGINE)
    store = engine.SessionStore
    if hasattr(store, "get_model_class"):
        session_model = store.get_model_class()
        if issubclass(session_model, AbstractBaseSession):
            return session_model._meta.app_config.name
    return None


def ensure_tenant_dict() -> None:
    if not isinstance(getattr(settings, "TENANTS", None), dict):
        raise ImproperlyConfigured("TENANTS dict setting not set.")


def ensure_public_schema() -> None:
    if not isinstance(settings.TENANTS.get("public"), dict):
        raise ImproperlyConfigured("TENANTS must contain a 'public' dict.")

    tenants_public = settings.TENANTS["public"]

    if "URLCONF" in tenants_public:
        raise ImproperlyConfigured("TENANTS['public'] cannot contain a 'URLCONF' key.")
    if "WS_URLCONF" in tenants_public:
        raise ImproperlyConfigured("TENANTS['public'] cannot contain a 'WS_URLCONF' key.")
    if "DOMAINS" in tenants_public:
        raise ImproperlyConfigured("TENANTS['public'] cannot contain a 'DOMAINS' key.")
    if "SESSION_KEY" in tenants_public:
        raise ImproperlyConfigured("TENANTS['public'] cannot contain a 'SESSION_KEY' key.")
    if "HEADER" in tenants_public:
        raise ImproperlyConfigured("TENANTS['public'] cannot contain a 'HEADER' key.")
    if "FALLBACK_DOMAINS" in tenants_public:
        raise ImproperlyConfigured("TENANTS['public'] cannot contain a 'FALLBACK_DOMAINS' key.")


def ensure_default_schemas() -> None:
    if "default" not in settings.TENANTS:
        return  # Escape hatch for static only configs

    if not isinstance(settings.TENANTS["default"], dict):
        raise ImproperlyConfigured("TENANTS must contain a 'default' dict.")

    tenants_default = settings.TENANTS["default"]

    if "TENANT_MODEL" not in tenants_default:
        raise ImproperlyConfigured("TENANTS['default'] must contain a 'TENANT_MODEL' key.")
    if "URLCONF" not in tenants_default:
        raise ImproperlyConfigured("TENANTS['default'] must contain a 'URLCONF' key.")
    if "DOMAINS" in tenants_default:
        raise ImproperlyConfigured("TENANTS['default'] cannot contain a 'DOMAINS' key.")
    if "SESSION_KEY" in tenants_default:
        raise ImproperlyConfigured("TENANTS['default'] cannot contain a 'SESSION_KEY' key.")
    if "HEADER" in tenants_default:
        raise ImproperlyConfigured("TENANTS['default'] cannot contain a 'HEADER' key.")
    if "FALLBACK_DOMAINS" in tenants_default:
        raise ImproperlyConfigured("TENANTS['default'] cannot contain a 'FALLBACK_DOMAINS' key.")
    if tenants_default.get("CLONE_REFERENCE") in settings.TENANTS:
        raise ImproperlyConfigured(
            "TENANTS['default']['CLONE_REFERENCE'] must be a unique schema name."
        )


def ensure_overall_schemas() -> None:
    for schema in settings.TENANTS:
        if schema not in ["public", "default"]:
            if not is_valid_schema_name(schema):
                raise ImproperlyConfigured(f"'{schema}' is not a valid schema name.")


def ensure_extra_search_paths() -> None:
    if not (extra_search_paths := get_extra_search_paths()):
        return

    TenantModel = get_tenant_model()

    dynamic_tenants = []

    if "default" in settings.TENANTS and "CLONE_REFERENCE" in settings.TENANTS["default"]:
        dynamic_tenants.append(settings.TENANTS["default"]["CLONE_REFERENCE"])

    if TenantModel is not None:
        with connection.cursor() as cursor:
            cursor.execute(
                "SELECT 1 FROM information_schema.tables WHERE table_name = %s;",
                [TenantModel._meta.db_table],
            )
            if cursor.fetchone():
                dynamic_tenants += list(TenantModel.objects.values_list("schema_name", flat=True))

    invalid_schemas = set(extra_search_paths) & (
        set(settings.TENANTS.keys()) | set(dynamic_tenants)
    )

    if invalid_schemas:
        invalid = ", ".join(invalid_schemas)
        raise ImproperlyConfigured(f"Do not include '{invalid}' on PGSCHEMAS_EXTRA_SEARCH_PATHS.")


@checks.register()
def check_principal_apps(app_configs: Any, **kwargs: Any) -> list:
    errors = []
    tenant_app = get_tenant_app()
    domain_app = get_domain_app()

    tenants_public = settings.TENANTS["public"]

    if tenant_app is not None and tenant_app not in tenants_public.get("APPS", []):
        errors.append(
            checks.Error(
                f"Your tenant app '{tenant_app}' must be on the 'public' schema.",
                id="pgschemas.W001",
            )
        )
    if domain_app is not None and domain_app not in tenants_public.get("APPS", []):
        errors.append(
            checks.Error(
                f"Your domain app '{domain_app}' must be on the 'public' schema.",
                id="pgschemas.W001",
            )
        )

    for schema in settings.TENANTS:
        schema_apps = settings.TENANTS[schema].get("APPS", [])
        if schema == "public":
            continue
        if tenant_app is not None and tenant_app in schema_apps:
            errors.append(
                checks.Error(
                    f"Your tenant app '{tenant_app}' in TENANTS['{schema}']['APPS'] "
                    "must be on the 'public' schema only.",
                    id="pgschemas.W001",
                )
            )
        if domain_app is not None and domain_app in schema_apps:
            errors.append(
                checks.Error(
                    f"Your domain app '{domain_app}' in TENANTS['{schema}']['APPS'] "
                    "must be on the 'public' schema only.",
                    id="pgschemas.W001",
                )
            )

    return errors


@checks.register()
def check_other_apps(app_configs: Any, **kwargs: Any) -> list:
    errors = []
    user_app = get_user_app()
    session_app = get_session_app()

    if "django.contrib.contenttypes" in settings.TENANTS.get("default", {}).get("APPS", []):
        errors.append(
            checks.Warning(
                "'django.contrib.contenttypes' in TENANTS['default']['APPS'] "
                "must be on 'public' schema only.",
                id="pgschemas.W002",
            )
        )

    for schema in settings.TENANTS:
        schema_apps = settings.TENANTS[schema].get("APPS", [])
        if schema not in ["public", "default"]:
            if "django.contrib.contenttypes" in schema_apps:
                errors.append(
                    checks.Warning(
                        f"'django.contrib.contenttypes' in TENANTS['{schema}']['APPS'] "
                        "must be on 'public' schema only.",
                        id="pgschemas.W002",
                    )
                )
        if user_app and session_app:
            if session_app in schema_apps and user_app not in schema_apps:
                errors.append(
                    checks.Warning(
                        f"'{user_app}' must be together with '{session_app}' in "
                        f"TENANTS['{schema}']['APPS'].",
                        id="pgschemas.W003",
                    )
                )
            elif (
                user_app in schema_apps
                and session_app not in schema_apps
                and session_app in settings.INSTALLED_APPS
            ):
                errors.append(
                    checks.Warning(
                        f"'{session_app}' must be together with '{user_app}' in "
                        f"TENANTS['{schema}']['APPS'].",
                        id="pgschemas.W003",
                    )
                )

    return errors


@checks.register(checks.Tags.database)
def check_schema_names(app_configs: Any, **kwargs: Any) -> list:
    errors = []
    static_names = set(settings.TENANTS.keys())
    clone_reference = get_clone_reference()
    TenantModel = get_tenant_model()

    if TenantModel is None:
        return []

    if clone_reference:
        static_names.add(clone_reference)
    try:
        dynamic_names = set(TenantModel.objects.values_list("schema_name", flat=True))
    except ProgrammingError:
        # This happens on the first run of migrate, with empty database.
        # It can also happen when the tenant model contains unapplied migrations that break.
        dynamic_names = set()

    intersection = static_names & dynamic_names

    if intersection:
        errors.append(
            checks.Critical(
                f"Name clash found between static and dynamic tenants: {intersection}",
                id="pgschemas.W004",
            )
        )

    return errors



import logging
from typing import Any

from django_pgschemas.routing.info import DomainInfo, HeadersInfo, SessionInfo
from django_pgschemas.schema import get_current_schema


class SchemaContextFilter(logging.Filter):
    """
    Add the current routing info to log records.
    """

    def filter(self, record: Any) -> bool:
        current_schema = get_current_schema()
        record.schema_name = current_schema.schema_name

        match current_schema.routing:
            case DomainInfo(domain, folder):
                record.domain = domain
                record.folder = folder
            case SessionInfo(reference):
                record.reference = reference
            case HeadersInfo(reference):
                record.reference = reference
            case _:
                pass

        return True



from django.apps import AppConfig as BaseAppConfig


class AppConfig(BaseAppConfig):
    name = "django_pgschemas"

    def ready(self) -> None:
        from . import checks  # noqa
        from .checks import (
            ensure_tenant_dict,
            ensure_public_schema,
            ensure_default_schemas,
            ensure_overall_schemas,
            ensure_extra_search_paths,
        )

        ensure_tenant_dict()
        ensure_public_schema()
        ensure_default_schemas()
        ensure_overall_schemas()
        ensure_extra_search_paths()



import re
from typing import Callable, TypeAlias, cast

from asgiref.sync import iscoroutinefunction, sync_to_async
from django.conf import settings
from django.db.models import Q
from django.http import Http404, HttpRequest, HttpResponse
from django.shortcuts import redirect
from django.urls import clear_url_caches, set_urlconf
from django.utils.decorators import sync_and_async_middleware

from django_pgschemas.models import TenantModel as TenantModelBase
from django_pgschemas.routing.info import DomainInfo, HeadersInfo, SessionInfo
from django_pgschemas.routing.models import get_primary_domain_for_tenant
from django_pgschemas.routing.urlresolvers import get_urlconf_from_schema
from django_pgschemas.schema import Schema, activate, activate_public
from django_pgschemas.settings import get_tenant_header, get_tenant_session_key
from django_pgschemas.utils import get_domain_model, get_tenant_model, remove_www


def strip_tenant_from_path_factory(prefix: str) -> Callable[[str], str]:
    def strip_tenant_from_path(path: str) -> str:
        return re.sub(r"^/{}/".format(prefix), "/", path)

    return strip_tenant_from_path


ResponseHandler: TypeAlias = Callable[[HttpRequest], HttpResponse]


def route_domain(request: HttpRequest) -> HttpResponse | None:
    hostname = remove_www(request.get_host().split(":")[0])

    activate_public()
    tenant: Schema | None = None

    # Checking for static tenants
    for schema, data in settings.TENANTS.items():
        if schema in ["public", "default"]:
            continue
        if hostname in data.get("DOMAINS", []):
            tenant = Schema.create(
                schema_name=schema,
                routing=DomainInfo(domain=hostname),
            )
            break

    # Checking for dynamic tenants
    else:
        DomainModel = get_domain_model()

        prefix = request.path.split("/")[1]
        domain = None

        if DomainModel is not None:
            try:
                domain = DomainModel.objects.select_related("tenant").get(
                    domain=hostname, folder=prefix
                )
            except DomainModel.DoesNotExist:
                try:
                    domain = DomainModel.objects.select_related("tenant").get(
                        domain=hostname, folder=""
                    )
                except DomainModel.DoesNotExist:
                    pass

        if domain is not None:
            tenant = cast(TenantModelBase, domain.tenant)
            tenant.routing = DomainInfo(domain=hostname)
            request.strip_tenant_from_path = lambda x: x

            if prefix and domain.folder == prefix:
                tenant.routing = DomainInfo(domain=hostname, folder=prefix)
                request.strip_tenant_from_path = strip_tenant_from_path_factory(prefix)
                clear_url_caches()  # Required to remove previous tenant prefix from cache (#8)

            if domain.redirect_to_primary:
                primary_domain = get_primary_domain_for_tenant(tenant)
                if primary_domain:
                    path = request.strip_tenant_from_path(request.path)
                    return redirect(primary_domain.absolute_url(path), permanent=True)

    # Checking fallback domains
    if not tenant:
        for schema, data in settings.TENANTS.items():
            if schema in ["public", "default"]:
                continue
            if hostname in data.get("FALLBACK_DOMAINS", []):
                tenant = Schema.create(
                    schema_name=schema,
                    routing=DomainInfo(domain=hostname),
                )
                break

    # No tenant found from domain / folder
    if not tenant:
        raise Http404("No tenant for hostname '%s'" % hostname)

    urlconf = get_urlconf_from_schema(tenant)

    request.tenant = tenant
    request.urlconf = urlconf
    set_urlconf(urlconf)

    activate(tenant)
    return None


def route_session(request: HttpRequest) -> HttpResponse | None:
    tenant_session_key = get_tenant_session_key()

    if not hasattr(request, "session") or not (
        tenant_ref := request.session.get(tenant_session_key)
    ):
        return None

    tenant: Schema | None = None

    # Checking for static tenants
    for schema, data in settings.TENANTS.items():
        if schema in ["public", "default"]:
            continue
        if tenant_ref == schema or tenant_ref == data.get("SESSION_KEY"):
            tenant = Schema.create(schema_name=schema)
            break

    # Checking for dynamic tenants
    else:
        if (TenantModel := get_tenant_model()) is not None:
            tenant = TenantModel._default_manager.filter(
                Q(pk__iexact=tenant_ref) | Q(schema_name=tenant_ref)
            ).first()

    if tenant is not None:
        tenant.routing = SessionInfo(reference=tenant_ref)
        request.tenant = tenant
        activate(tenant)

    return None


def route_headers(request: HttpRequest) -> HttpResponse | None:
    tenant_header = get_tenant_header()

    if not (tenant_ref := request.headers.get(tenant_header)):
        return None

    tenant: Schema | None = None

    # Checking for static tenants
    for schema, data in settings.TENANTS.items():
        if schema in ["public", "default"]:
            continue
        if tenant_ref == schema or tenant_ref == data.get("HEADER"):
            tenant = Schema.create(schema_name=schema)
            break

    # Checking for dynamic tenants
    else:
        if (TenantModel := get_tenant_model()) is not None:
            tenant = TenantModel._default_manager.filter(
                Q(pk__iexact=tenant_ref) | Q(schema_name=tenant_ref)
            ).first()

    if tenant is not None:
        tenant.routing = HeadersInfo(reference=tenant_ref)
        request.tenant = tenant
        activate(tenant)

    return None


def middleware_factory(
    handler: Callable[[HttpRequest], HttpResponse | None],
) -> Callable[[ResponseHandler], ResponseHandler]:
    @sync_and_async_middleware
    def middleware(get_response: ResponseHandler) -> ResponseHandler:
        if iscoroutinefunction(get_response):
            async_base_middleware = sync_to_async(handler)

            async def sync_middleware(request: HttpRequest) -> HttpResponse | None:
                if response := await async_base_middleware(request):
                    return response

                return await get_response(request)

            return sync_middleware

        else:

            def async_middleware(request: HttpRequest) -> HttpResponse | None:
                if response := handler(request):
                    return response

                return get_response(request)

            return async_middleware

    return middleware


DomainRoutingMiddleware = middleware_factory(route_domain)
SessionRoutingMiddleware = middleware_factory(route_session)
HeadersRoutingMiddleware = middleware_factory(route_headers)
